#property copyright "Copyright 2022, Anil Varma"
#property link      "AnilHVarma@gmail.com"
#property version   "1.00"

struct SiStoch {
	double main;
	double signal;
};

#include "../MQL5 Forum/IndicatorBaseMQL5.mqh"
//+-----------------------------------------------------------------------------------------------------------------------------+
//| CLASS:				CStochMQL5.mqh
//| APPLICATION:	CLASS FOR STOCHASTIC OSCILLATOR
//+-----------------------------------------------------------------------------------------------------------------------------+
class CStochMQL5 final : public CIndicatorBaseMQL5 {

private:

  int                 mHandle;
	int									mKPeriod;									// STOCHASTIC OSCILLATOR %K PERIOD
	int									mDPeriod;									// STOCHASTIC OSCILLATOR %D PERIOD
	int									mSlowing;									// STOCHASTIC OSCILLATOR SLOWING PERIOD
	ENUM_STO_PRICE			mAppliedPriceStoch;				// STOCHASTIC OSCILLATOR APPLIED PRICE

	// SETTER METHOD FOR VARIABLES
	int									mThldOverBought;
	int									mThldOverSold;

public:

  CStochMQL5() : CIndicatorBaseMQL5() { }
  CStochMQL5(string pSymbol,ENUM_TIMEFRAMES pTimeFrame,int pKPeriod,int pDPeriod,int pSlowing,int pThldOB=80)
  	: CIndicatorBaseMQL5() { InitHandle(pSymbol,pTimeFrame,pKPeriod,pDPeriod,pSlowing,pThldOB); }
 ~CStochMQL5();

	// METHOD OF INITIALIZE INDICATOR HANDLES
	void				InitHandle(string pSymbol,ENUM_TIMEFRAMES pTimeFrame,int pKPeriod,int pDPeriod,int pSlowing,int pThldOB=80);

	// METHODS TO RETURN INDICATOR VALUES TO CALLING EA / CLASS
	void 				GetArrayStoch(int pCount,SiStoch &pStoch[]);
	double			GetIndexMain(int pIndex);
	double			GetIndexSignal(int pIndex);

	bool 				IsStochCrossLONG(int pIndex);
	bool 				IsStochCrossSHORT(int pIndex);

	// METHOD TO CHECK DIVERGENCE
	bool 				IsClassicDVRG_BEARISH(int pIndexB,int pIndexA);
	bool 				IsClassicDVRG_BULLISH(int pIndexB,int pIndexA);
	bool 				IsHiddenDVRG_BEARISH(int pIndexB,int pIndexA);
	bool 				IsHiddenDVRG_BULLISH(int pIndexB,int pIndexA);

	// METHOD TO CHECK CONVERGENCE
	bool 				IsCVRG_BULLISH(int pIndexB,int pIndexA);
	bool 				IsCVRG_BEARISH(int pIndexB,int pIndexA);


}; // END Of Class Definition
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD:       ~CStochMQL5()
//| APPLICATION:  DECONSTRUCTOR METHOD
//+-----------------------------------------------------------------------------------------------------------------------------+
CStochMQL5::~CStochMQL5() {

  IndicatorRelease(mHandle);

} // END Of default deconstructor ~CStochMQL5()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD:				InitHandle()
//| APPLICATION:	method to intiliaze member variables from input parameters
//+-----------------------------------------------------------------------------------------------------------------------------+
void CStochMQL5::InitHandle(string pSymbol,ENUM_TIMEFRAMES pTimeFrame,int pKPeriod,int pDPeriod,int pSlowing,int pThldOB=80) {

	// ASSIGN SYMBOL AND TIME FRAME
	mSymbol		 = pSymbol;
	mTimeFrame = pTimeFrame;

	// ASSIGN PARAMETERS TO LOCAL VARIABLES
	mKPeriod			 		 = pKPeriod;
	mDPeriod		 	 		 = pDPeriod;
	mSlowing 					 = pSlowing;
	mAppliedPriceStoch = STO_LOWHIGH;								// HARD CODED THE STOCHASTIC PRICE TYPE FROM 'ENUM_STO_PRICE'

	mThldOverBought		 = pThldOB;
	mThldOverSold			 = 100 - mThldOverBought;

	// DECLARE INDICATOR HANDLES
	TesterHideIndicators(true);

	// DEFAULT VALUES OF CIndicatorBaseMQL5 mAppliedMAMode = MODE_LWMA IS APPLIED
	mHandle = iStochastic(mSymbol,mTimeFrame,mKPeriod,mDPeriod,mSlowing,mAppliedMAMode,mAppliedPriceStoch);
  if(mHandle == INVALID_HANDLE) {
		ResetLastError();
    PrintFormat("[%s] %s: Error[#%i] creating handle iStochastic",EnumToString(mTimeFrame),(string)__FUNCTION__,GetLastError());
  }

	TesterHideIndicators(false);

} // END Of method Init()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD:  			GetArrayStoch()
//| APPLICATION:  METHOD TO RETURN iDBB VALUES INTO PASSED IN STRUCTIRE ARRAY
//+-----------------------------------------------------------------------------------------------------------------------------+
void CStochMQL5::GetArrayStoch(int pCount,SiStoch &pStoch[]) {

  int start = 0;            												// START DEFAULTED TO INDEX[0] TO ALLIGN BARS WITH CHART
  int count = pCount + 1;   												// EXTRA BAR ADDED FOR INDEX[0]

	ArraySetAsSeries(pStoch,true);
	ArrayResize(pStoch,count);

	// MAIN LOOP TO GET VALUES
  for(int i = count - 1; i >= start; i--) {
	  pStoch[i].main	 = GetIndexMain(i);
	  pStoch[i].signal = GetIndexSignal(i);
  }

} // END Of method GetArrayStoch()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD:				GetIndexMain()
//| APPLICATION:	RETURNS STOCHASTIC MAIN_LINE VALUE
//+-----------------------------------------------------------------------------------------------------------------------------+
double CStochMQL5::GetIndexMain(int pIndex) {

	// IF HANDLES NOT VALID, REINITIALIZE THEM
  if(mHandle == INVALID_HANDLE) {
    InitHandle(mSymbol,mTimeFrame,mKPeriod,mDPeriod,mSlowing);
  }

  double arrayData[1];       												// LOCAL STATIC BUFFER TO RETURN SINGLE INDEX VALUE

	// START POSITION IS pIndex AND COUNT IS SINGLE VALUE
  if(CopyBuffer(mHandle,MAIN_LINE,pIndex,1,arrayData) < 1) {
		ResetLastError();
    PrintFormat("[%s] %s: Error[#%i] getting iStochastic MAIN at Index[%i]",EnumToString(mTimeFrame),string(__FUNCTION__),
    						GetLastError(),pIndex);
    return(0.00);
  }

  return(arrayData[0]);

} // END Of method GetIndexMain()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD:				GetIndexSignal()
//| APPLICATION:	RETURNS AVERAGE TRUE RANGE (ATR) VALUE
//+-----------------------------------------------------------------------------------------------------------------------------+
double CStochMQL5::GetIndexSignal(int pIndex) {

	// IF HANDLES NOT VALID, REINITIALIZE THEM
  if(mHandle == INVALID_HANDLE) {
    InitHandle(mSymbol,mTimeFrame,mKPeriod,mDPeriod,mSlowing);
  }

  double arrayData[1];       												// LOCAL STATIC BUFFER TO RETURN SINGLE INDEX VALUE

	// START POSITION IS pIndex AND COUNT IS SINGLE VALUE
  if(CopyBuffer(mHandle,SIGNAL_LINE,pIndex,1,arrayData) < 1) {
		ResetLastError();
    PrintFormat("[%s] %s: Error[#%i] getting iStochastic SIGNAL at Index[%i]",EnumToString(mTimeFrame),string(__FUNCTION__),
    						GetLastError(),pIndex);
    return(0.00);
  }

  return(arrayData[0]);

} // END Of method GetIndexSignal()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD BODY:  IsClassicDVRG_BEARISH()
//| APPLICATION:  Classic BEARISH Divergence: 			Price: HIGHER HIGH 	and MFI: LOWER HIGH
//+-----------------------------------------------------------------------------------------------------------------------------+
bool CStochMQL5::IsClassicDVRG_BEARISH(int pIndexB,int pIndexA) {

	// DEFINE VARIABLE
	double idxB_HIGH = iHigh(mSymbol,mTimeFrame,pIndexB);
	double idxA_HIGH = iHigh(mSymbol,mTimeFrame,pIndexA);

	if(idxB_HIGH < idxA_HIGH && GetIndexMain(pIndexB) > GetIndexMain(pIndexA)) 		return(true);
	else 																																					return(false);

} // END Of method IsClassicDVRG_BEARISH()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD BODY:  IsClassicDVRG_BULLISH()
//| APPLICATION:  Classic BULLISH Divergence:				Price: LOWER LOW 	 	and MFI: HIGHER LOW
//+-----------------------------------------------------------------------------------------------------------------------------+
bool CStochMQL5::IsClassicDVRG_BULLISH(int pIndexB,int pIndexA) {

	// DEFINE VARIABLE
	double idxB_LOW  = iLow(mSymbol,mTimeFrame,pIndexB);
	double idxA_LOW  = iLow(mSymbol,mTimeFrame,pIndexA);

	if(idxB_LOW > idxA_LOW && GetIndexMain(pIndexB) < GetIndexMain(pIndexA)) 			return(true);
	else 																																					return(false);

} // END Of method IsClassicDVRG_BULLISH()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD BODY:  IsHiddenDVRG_BEARISH()
//| APPLICATION:  Hidden 	BEARISH Divergence: 			Price: LOWER HIGH 	and MFI: HIGHER HIGH
//| NOTE:					STOCHASING AppliedPrice IS STO_LOWHIGH, HENCE WE USED NORMAL HIGH / LOW PRICE
//+-----------------------------------------------------------------------------------------------------------------------------+
bool CStochMQL5::IsHiddenDVRG_BEARISH(int pIndexB,int pIndexA) {

	// DEFINE VARIABLE
	double idxB_HIGH = iHigh(mSymbol,mTimeFrame,pIndexB);
	double idxA_HIGH = iHigh(mSymbol,mTimeFrame,pIndexA);

	if(idxB_HIGH > idxA_HIGH && GetIndexMain(pIndexB) < GetIndexMain(pIndexA)) 		return(true);
	else 																																					return(false);

} // END Of method IsHiddenDVRG_BEARISH()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD BODY:  IsHiddenDVRG_BULLISH()
//| APPLICATION:  Hidden 	BULLISH Divergence:				Price: HIGHER LOW	 	and Oscillator: LOWER LOW
//| NOTE:					STOCHASING AppliedPrice IS STO_LOWHIGH, HENCE WE USED NORMAL HIGH / LOW PRICE
//+-----------------------------------------------------------------------------------------------------------------------------+
bool CStochMQL5::IsHiddenDVRG_BULLISH(int pIndexB,int pIndexA) {

	// DEFINE VARIABLE
	double idxB_LOW  = iLow(mSymbol,mTimeFrame,pIndexB);
	double idxA_LOW  = iLow(mSymbol,mTimeFrame,pIndexA);

	if(idxB_LOW < idxA_LOW && GetIndexMain(pIndexB) > GetIndexMain(pIndexA)) 			return(true);
	else 																																					return(false);

} // END Of method IsHiddenDVRG_BULLISH()
//+-----------------------------------------------------------------------------------------------------------------------------+
/*
	https://www.investopedia.com/ask/answers/121714/what-are-differences-between-divergence-and-convergence.asp
	Convergence occurs when the price of an asset, indicator, or index moves in the same direction as a related asset, indicator,
	or index in technical analysis.
*/
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD BODY:  IsCVRG_BULLISH()
//| APPLICATION:  BULLISH Divergence:				Price: HIGHER LOW and Oscillator: HIGHER LOW
//| NOTE:					WE USED NORMAL HIGH / LOW PRICE TO MAINTAIN COMPATIBILITY FOR MFI/STOCH DIVERGENCE
//+-----------------------------------------------------------------------------------------------------------------------------+
bool CStochMQL5::IsCVRG_BULLISH(int pIndexB,int pIndexA) {

	// DEFINE VARIABLE
	double idxB_LOW  = iLow(mSymbol,mTimeFrame,pIndexB);
	double idxA_LOW  = iLow(mSymbol,mTimeFrame,pIndexA);
/*
	PrintFormat("%s: [%s] IdxB[%i] IdxA[%i] Low[%.1f] < [%.1f] StochMain[%.5f] < [%.5f]",(string)__FUNCTION__,
							EnumToString(mTimeFrame),pIndexB,pIndexA,idxB_LOW,idxA_LOW,GetIndexMain(pIndexB),GetIndexMain(pIndexA));
*/

	if(idxB_LOW < idxA_LOW && GetIndexMain(pIndexB) < GetIndexMain(pIndexA)) 		return(true);
	else 																																				return(false);

} // END Of method IsCVRG_BULLISH()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD BODY:  IsCVRG_BEARISH()
//| APPLICATION:  BEARISH Divergence:				Price: LOWER HIGH and Oscillator: LOWER HIGH
//| NOTE:					WE USED NORMAL HIGH / LOW PRICE TO MAINTAIN COMPATIBILITY FOR MFI/STOCH DIVERGENCE
//+-----------------------------------------------------------------------------------------------------------------------------+
bool CStochMQL5::IsCVRG_BEARISH(int pIndexB,int pIndexA) {

	// DEFINE VARIABLE
	double idxB_HIGH = iHigh(mSymbol,mTimeFrame,pIndexB);
	double idxA_HIGH = iHigh(mSymbol,mTimeFrame,pIndexA);

	if(idxB_HIGH > idxA_HIGH && GetIndexMain(pIndexB) > GetIndexMain(pIndexA)) 	return(true);
	else 																																				return(false);

} // END Of method IsCVRG_BULLISH()
//+-----------------------------------------------------------------------------------------------------------------------------+
